plus:: Nat ~> Nat ~> Nat
{plus Z m} = m
{plus (S n) m} = S {plus n m}
  
data Seq:: *0 ~> Nat ~> *0 where
  Snil :: Seq a Z
  Scons:: a -> Seq a n -> Seq a (S n)
  
l1 = (Scons 3 (Scons 5 Snil)) :: Seq Int (S(S Z))

app:: Seq a n -> Seq a m -> Seq a {plus n m}
app Snil ys = ys
app (Scons x xs) ys = Scons x (app xs ys)

length:: Seq a n -> Nat' n
length Snil = Z
length (Scons x xs) = S(length xs)

--------------------------------------

data Boolean:: *1 where
  T:: Boolean
  F:: Boolean  
  
 
even :: Nat ~> Boolean
{even Z} = T
{even (S Z)} = F
{even (S (S n))} = {even n}


le:: Nat ~> Nat ~> Boolean 
{le Z n} = T
{le (S n) Z} = F
{le (S n) (S m)} = {le n m} 

isEven:: Nat' n -> Maybe(Proof {even n})
isEven Z = Just Triv
isEven (S Z) = Nothing
isEven (S (S n)) = isEven n

and:: Boolean ~> Boolean ~> Boolean
{and T x} = x
{and F x} = F

##test "illegal proof by type"
 bad:: Proof {even 1t}
 bad = Triv


smaller :: Proof {le (S a) (S b)} -> Proof {le a b}
smaller Triv = Triv

diff:: Proof {le a b} -> Nat' a -> Nat' b -> exists c .(Nat' c,Equal {plus a c} b)
diff Triv Z m = Ex (m,Eq)
diff Triv (S m) Z = unreachable
diff (q@Triv) (S x) (S y) =  
  case diff (smaller q) x y of
   Ex (m,Eq) -> Ex (m,Eq)
      
--
--------------------------------------
-- data Tree a = Fork (Tree a) (Tree a) | Node a | Tip

data HTree a = HFork (HTree a) (HTree a) | HNode a | HTip

data Shape :: *1 where 
  Tp:: Shape
  Nd:: Shape
  Fk:: Shape ~> Shape ~> Shape
 
data Path:: Shape ~> *0 ~> *0 where
  None :: Path Tp a
  Here :: b -> Path Nd b
  Left :: Path x a -> Path (Fk x y) a
  Right:: Path y a -> Path (Fk x y) a
  
data Tree :: Shape ~> *0 ~> *0 where
  Tip:: Tree Tp a
  Node:: a -> Tree Nd a
  Fork::  Tree x a -> Tree y a -> Tree (Fk x y) a

find:: (a -> a -> Bool) -> a -> Tree s a -> [Path s a]
find eq n Tip = []
find eq n (Node m) =
  if eq n m then [Here n] else []
find eq n (Fork x y) = 
  map Left (find eq n x) ++ 
  map Right (find eq n y)

tree1 :: Tree (Fk (Fk Tp Nd) (Fk Nd Nd)) Int
tree1 = Fork (Fork Tip (Node 4)) (Fork (Node 4) (Node 3))

map f [] = []
map f (x:xs) = f x : map f xs

----------------------------------------------------

data Proof:: Boolean ~> *0 where
  Triv:: Proof T

data Plus:: Nat ~> Nat ~> Nat ~> *0 
      where
  PlusZ:: Plus Z m m
  PlusS:: Plus n m z -> 
          Plus (S n) m (S z)

data LE:: Nat ~> Nat ~> *0 where
  LeZ:: LE Z n
  LeS:: LE n m -> 
        LE (S n) (S m)

data Even:: Nat ~> *0 where
  EvenZ:: Even Z
  EvenSS:: Even n -> Even (S (S n))

even0:: Even 0t
even0 = EvenZ

even2 :: Even 2t
even2 = (EvenSS EvenZ)

even4 :: Even 4t
even4 = (EvenSS (EvenSS EvenZ))

less0:: LE 0t a
less0 = LeZ 
 
less1:: LE 1t (1+a)t 
less1 = (LeS LeZ)
 
less2 :: LE 2t (2+a)t 
less2 =  (LeS (LeS LeZ))
 
even2p :: Proof {even 2t}
even2p = Triv 

{-

-}
--------------------------------

data Exp:: *0 where
  Variable:: String  -> Exp
  Constant:: Int -> Exp
  Plus:: Exp
  Less:: Exp
  Apply:: Exp -> Exp -> Exp
  Tuple:: [Exp] -> Exp 

exp1 = Apply Plus (Tuple [Variable "x" ,Variable "y"])

data Value :: *0 where
  IntV:: Int -> Value
  BoolV:: Bool -> Value
  FunV:: (Value -> Value) -> Value
  TupleV :: [Value] -> Value
  
eval:: (String -> Value) -> Exp -> Value
eval env (Variable s) = env s
eval env (Constant n) = IntV n
eval env Plus = FunV plus
  where plus (TupleV[IntV n ,IntV m]) = IntV(n+m)
eval env Less = FunV less
  where less (TupleV[IntV n ,IntV m]) = BoolV(n < m) 
eval env (Apply f x) = 
  case eval env f of
    FunV g -> g (eval env x)
eval env (Tuple xs) = TupleV(map (eval env) xs) 

stagedEval:: (String -> Code Value) -> Exp -> Code Value
stagedEval env (Variable s) = env s
stagedEval env (Constant n) = lift(IntV n)
stagedEval env Plus = [| FunV plus |]
  where plus (TupleV[IntV n ,IntV m]) = IntV(n+m)
stagedEval env Less = [| FunV less |]
  where less (TupleV[IntV n ,IntV m]) = BoolV(n < m) 
stagedEval env (Apply f x) = [| apply $(stagedEval env f) $(stagedEval env x) |]
  where apply (FunV g) x = g x
stagedEval env (Tuple xs) = [| TupleV $(mapLift (stagedEval env) xs) |]
  where mapLift f [] = lift []
        mapLift f (x:xs) = [| $(f x) : $(mapLift f xs) |]
        
ans = stagedEval f exp1 
  where f "x" = lift(IntV 3)
        f "y" = lift(IntV 4)
        
-- [| %apply (%FunV %plus) (%TupleV [IntV 3,IntV 4]) |] : Code Value   

------------------------------------------

data Term:: *0 ~> *0 where
  Const :: Int -> Term Int               -- 5
  Add:: Term ((Int,Int) -> Int)          -- (+)
  LT:: Term ((Int,Int) -> Bool)          -- (<)
  Ap:: Term(a -> b) -> Term a -> Term b  -- (+) (x,y)
  Pair:: Term a -> Term b -> Term(a,b)   -- (x,y)
  
ex1 :: Term Int  
ex1 = Ap Add (Pair (Const 3) (Const 5))

ex2 :: Term (Int,Int)
ex2 = Pair ex1 (Const 1) 

evalTerm :: Term a -> a
evalTerm (Const x) = x
evalTerm Add = \ (x,y) -> x+y
evalTerm LT = \ (x,y) -> x<y
evalTerm (Ap f x) = evalTerm f (evalTerm x)
evalTerm (Pair x y) = (evalTerm x,evalTerm y)

stagedEvalTerm :: Term a -> Code a
stagedEvalTerm (Const x) = lift x
stagedEvalTerm Add = [| add |]
  where add (x,y) = x+y
stagedEvalTerm LT = [| less |]
  where less (x,y) = x < y
stagedEvalTerm (Ap f x) = [| $(stagedEvalTerm f) $(stagedEvalTerm x) |]
stagedEvalTerm (Pair x y) = [| ( $(stagedEvalTerm x),$(stagedEvalTerm y)) |]

ans2 = stagedEvalTerm ex2

-- ans2 = [| (%add (3, 5), 1) |] : Code (Int,Int)

---------------------------------------------------

data Rep :: *0 ~> *0 where
   Int :: Rep Int
   Prod :: Rep a -> Rep b -> Rep (a,b)
   List :: Rep a -> Rep [a]

same:: Nat' n -> LE n n
same Z = LeZ
same (S n) = LeS(same n)

predLE:: Nat' n -> LE n (S n)
predLE Z = LeZ
predLE (S n) = LeS(predLE n)

trans:: LE a b -> LE b c -> LE a c
trans LeZ LeZ = LeZ
trans LeZ (LeS m) = LeZ
trans (LeS n) LeZ = unreachable
trans (LeS n) (LeS m) = LeS(trans n m)

summandLessThanSum:: Plus a b c -> LE a c
summandLessThanSum PlusZ = LeZ
summandLessThanSum (PlusS n) =  LeS(summandLessThanSum n)



ff:: Nat' b -> Plus a b c -> LE b c
ff Z _ = LeZ
ff (S n) PlusZ = LeS(ff n PlusZ)
ff (S n) (PlusS m) = LeS(trans (predLE n) (ff (S n) m))

filter :: (a->Bool) -> Seq a n -> exists m . (LE m n,Nat' m,Seq a m)
filter p Snil = Ex(LeZ,Z,Snil)
filter p (Scons x xs) =
  case p x of
    True  -> case filter p xs of
              Ex (p,n,ys) -> Ex(LeS p,S n,Scons x ys)
    False -> case filter p xs of
              Ex(p,n,ys) -> Ex (trans (predLE n) (LeS p),n,ys)


filter2:: (a -> Bool) -> Seq a n -> exists i j . (Seq a i,Nat' j,Equal n {plus j i})
filter2 p Snil = Ex(Snil,Z,Eq)
filter2 p (Scons x xs) = 
  case (filter2 p xs,p x) of
    -- (Ex(ys,n,Eq),True) -> Ex(Scons x ys,n,Eq)
    -- needs plus commutes
    (Ex(ys,n,Eq),False) -> Ex(ys,(S n),Eq)
    
pow:: Int -> Code Int -> Code Int
pow 0 x = [| 1 |]
pow n x = [| $x * $(pow (n-1) x) |]

pow4 = pow 4 [|99|]

data Nsum:: *0 ~> *0 where
  SumZ:: Nsum Int
  SumS:: Nsum x -> Nsum (Int -> x)
 deriving Nat(i)  
 
add:: Nsum i -> i
add 0i = 0
add 1i = \ x -> x
add (2+n)i = \ y -> \ x -> add (1+n)i (y+x)


data Expr:: *0 where
  VarExpr :: Label t -> Expr
  PlusExpr:: Expr -> Expr -> Expr

lookup:: Label v -> [exists t .(Label t,Int)] -> Int
lookup v ((Ex(u,n)):xs) = 
  case labelEq v u of
    Just Eq -> n
    Nothing -> lookup v xs
  
valueOf:: Expr -> [exists t .(Label t,Int)] -> Int
valueOf (VarExpr v) env = lookup v env
valueOf (PlusExpr x y) env = valueOf x env + valueOf y env
  

data VNum:: Tag ~> *0 ~> Row Tag *0 ~> *0 where
  Zv:: VNum l t (RCons l t row)
  Sv:: VNum l t (RCons a b row) -> VNum l t (RCons x y (RCons a b row))
 deriving Nat(u)
 
data Exp2:: Row Tag *0 ~> *0 ~> *0 where
  Var:: Label v -> VNum v t e -> Exp2 e t
  Greater:: Exp2 e Int -> Exp2 e Int -> Exp2 e Bool
  Sum:: Exp2 e Int -> Exp2 e Int -> Exp2 e Int
  If:: Exp2 e Bool -> Exp2 e t -> Exp2 e t -> Exp2 e t


sameNat:: Nat' a -> Nat' b -> Maybe(Equal a b)
sameNat (x@Z) (y@Z) = Just Eq
sameNat (x@(S n)) (y@(S m)) =
  case sameNat n m of
    Just Eq -> (Just Eq)
    
